#!/usr/bin/env python

from leather.axis import Axis
from leather.data_types import Number, Text
from leather.chart import Chart
from leather.grid import Grid
from leather.lattice import Lattice
from leather.scales import Scale, Linear, Ordinal, Temporal
from leather.series import Series, CategorySeries, key_function
from leather.shapes import Shape, Bars, Columns, Dots, Line, style_function
from leather.testcase import LeatherTestCase
from leather import theme



#!/usr/bin/env python

from collections import namedtuple
from datetime import date, datetime, timedelta
from decimal import Decimal
import math
import sys
import warnings

import six

try:
    __IPYTHON__
    from IPython.display import SVG as IPythonSVG
except (NameError, ImportError):
    IPythonSVG = lambda x: x


# Shorthand
ZERO = Decimal('0')
NINE_PLACES = Decimal('1e-9')

#: X data dimension index
X = 0

#: Y data dimension index
Y = 1

#: Z data dimension index
Z = 2


DIMENSION_NAMES = ['X', 'Y', 'Z']

#: Data structure for representing margins or other CSS-edge like properties
Box = namedtuple('Box', ['top', 'right', 'bottom', 'left'])

#: Data structure for a single series data point
Datum = namedtuple('Datum', ['i', 'x', 'y', 'z', 'row'])

#: Dummy object used in place of a series when rendering legends for categories
DummySeries = namedtuple('DummySeries', ['name'])


formatwarning_orig = warnings.formatwarning
warnings.formatwarning = lambda message, category, filename, lineno, line=None: \
    formatwarning_orig(message, category, filename, lineno, line='')

warn = warnings.warn
warnings.resetwarnings()
warnings.simplefilter('always')


# In Python 3.5 use builtin C implementation of `isclose`
if sys.version_info >= (3, 5):
    from math import isclose
else:
    def isclose(a, b, rel_tol=NINE_PLACES, abs_tol=ZERO):
        """
        Test if two floating points numbers are close enough to be considered
        equal.

        Via: https://github.com/PythonCHB/close_pep/blob/master/isclose.py

        Verified against final CPython 3.5 implementation.

        :param a:
            The first number to check.
        :param b:
            The second number to check.
        :param rel_tol:
            Relative tolerance. The amount of error allowed, relative to the larger
            input value. Defaults to nine decimal places of accuracy.
        :param abs_tol:
            Absolute minimum tolerance. Disabled by default.
        """
        if a == b:
            return True

        if rel_tol < ZERO or abs_tol < ZERO:
            raise ValueError('Tolerances must be non-negative')

        if math.isinf(abs(a)) or math.isinf(abs(b)):
            return False

        diff = abs(b - a)

        return (((diff <= abs(rel_tol * b)) or
                (diff <= abs(rel_tol * a))) or
                (diff <= abs_tol))


def to_year_count(d):
    """
    date > n years
    """
    return d.year


def from_year_count(n, t=date):
    """
    n years > date
    """
    return t(n, 1, 1)


def to_month_count(d):
    """
    date > n months
    """
    return (d.year * 12) + d.month


def from_month_count(n, t=date):
    """
    n months > date
    """
    return t(n // 12, (n % 12) + 1, 1)


def to_day_count(d):
    """
    date > n days
    """
    return (d - type(d).min).days


def from_day_count(n, t=date):
    """
    n days > date
    """
    return t.min + timedelta(days=n)


def to_hour_count(d):
    """
    date > n hours
    """
    return (d - datetime.min).total_seconds() / (60 * 60)


def from_hour_count(n, t=datetime):
    """
    n hours > date
    """
    return t.min + timedelta(hours=n)


def to_minute_count(d):
    """
    date > n minutes
    """
    return (d - datetime.min).total_seconds() / 60


def from_minute_count(n, t=datetime):
    """
    n minutes > date
    """
    return t.min + timedelta(minutes=n)


def to_second_count(d):
    """
    date > n seconds
    """
    return (d - datetime.min).total_seconds()


def from_second_count(n, t=datetime):
    """
    n seconds > date
    """
    return t.min + timedelta(seconds=n)


def to_microsecond_count(d):
    """
    date > n microseconds
    """
    return (d - datetime.min).total_seconds() * 1000


def from_microsecond_count(n, t=datetime):
    """
    n microseconds > date
    """
    return t.min + timedelta(microseconds=n)



#!/usr/bin/env python

from decimal import Decimal, ROUND_CEILING, ROUND_FLOOR
import math
import sys

from leather.ticks.base import Ticker
from leather.utils import isclose

# Shorthand
ZERO = Decimal('0')
TEN = Decimal('10')

#: Normalized intervals to be tested for ticks
INTERVALS = [
    Decimal('0.1'),
    Decimal('0.15'),
    Decimal('0.2'),
    Decimal('0.25'),
    Decimal('0.5'),
    Decimal('1.0')
]

#: The default number of ticks to produce
DEFAULT_TICKS = 5

#: The minimum length of a viable tick sequence
MIN_TICK_COUNT = 4

#: The maximum length of a viable tick sequence
MAX_TICK_COUNT = 10

#: Most preferred tick intervals
BEST_INTERVALS = [Decimal('0.1'), Decimal('1.0')]

#: Least preferred tick intervals
WORST_INTERVALS = [Decimal('0.15')]


class ScoreTicker(Ticker):
    """
    Attempt to find an optimal series of ticks by generating many possible
    sequences and scoring them based on several criteria. Only the best
    tick sequence is returned.

    Based an algorithm described by Austin Clemens:
    http://austinclemens.com/blog/2016/01/09/an-algorithm-for-creating-a-graphs-axes/

    See :meth:`.ScoreTicker.score` for scoring implementation.

    :param domain_min:
        Minimum value of the data series.
    :param domain_max:
        Maximum value of the data series.
    """
    def __init__(self, domain_min, domain_max):
        self._domain_min = domain_min
        self._domain_max = domain_max

        self._ticks = self._find_ticks()

        self._min = self._ticks[0]
        self._max = self._ticks[-1]

    @property
    def ticks(self):
        return self._ticks

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    def _find_ticks(self):
        """
        Implements the tick-finding algorithm.
        """
        force_zero = self._domain_min < ZERO and self._domain_max > ZERO

        interval_guess = abs(self._domain_max - self._domain_min) / (DEFAULT_TICKS - 1)
        magnitude = interval_guess.log10().to_integral_exact(rounding=ROUND_CEILING)

        candidate_intervals = []

        for interval in INTERVALS:
            candidate_intervals.append((interval, interval * pow(TEN, magnitude)))
            candidate_intervals.append((interval, interval * pow(TEN, magnitude - 1)))
            candidate_intervals.append((interval, interval * pow(TEN, magnitude + 1)))

        candidate_ticks = []

        for base_interval, interval in candidate_intervals:
            ticks = []

            if force_zero:
                min_steps = (abs(self._domain_min) / interval).to_integral_exact(rounding=ROUND_CEILING)
                ticks.append(self._round_tick(-min_steps * interval))
            else:
                ticks.append(self._round_tick((self._domain_min / interval).to_integral_exact(rounding=ROUND_FLOOR) * interval))

            tick_num = 1

            while ticks[tick_num - 1] < self._domain_max:
                t = self._round_tick(ticks[0] + (interval * tick_num))

                ticks.append(t)
                tick_num += 1

            # Throw out sequences that are too short or too long
            if len(ticks) < MIN_TICK_COUNT or len(ticks) > MAX_TICK_COUNT:
                continue

            candidate_ticks.append({
                'base_interval': base_interval,
                'interval': interval,
                'ticks': ticks,
                'score': self._score(base_interval, interval, ticks)
            })

        # Order by best score, using number of ticks as a tie-breaker
        best = sorted(candidate_ticks, key=lambda c: (c['score']['total'], len(c['ticks'])))

        return best[0]['ticks']

    def _score(self, base_interval, interval, ticks):
        """
        Score a given tick sequence based on several criteria. This method returns
        discrete scoring components for easier debugging.
        """
        s = {
            'pct_waste': 0,
            'interval_penalty': 0,
            'len_penalty': 0,
            'total': 0
        }

        # Penalty for wasted scale space
        waste = (self._domain_min - ticks[0]) + (ticks[-1] - self._domain_max)
        pct_waste = waste / (self._domain_max - self._domain_min)

        s['pct_waste'] = pow(10, pct_waste)

        # Penalty for choosing less optimal tick intervals
        if base_interval in BEST_INTERVALS:
            pass
        elif base_interval in WORST_INTERVALS:
            s['interval_penalty'] = 2
        else:
            s['interval_penalty'] = 1

        # Penalty for too many ticks
        if len(ticks) > 5:
            s['len_penalty'] = (len(ticks) - 5)

        s['total'] = s['pct_waste'] + s['interval_penalty'] + s['len_penalty']

        return s

    def _round_tick(self, t):
        """
        Round a tick to 0-3 decimal places, if the remaining digits do not
        appear to be significant.
        """
        for r in range(0, 4):
            exp = pow(Decimal(10), Decimal(-r))
            quantized = t.quantize(exp)

            if isclose(t, quantized):
                return quantized

        return t
